from typing import Tuple, Dict, List
import requests
from time import sleep
from tqdm import tqdm
import cv2
import random

from blueness import module

from blue_assistant import NAME
from blue_assistant import env
from blue_assistant.script.repository.hue.colors import get_hue_values
from blue_assistant.logger import logger

NAME = module.name(__file__, NAME)


# hue-2025-03-14-1l8tv6
def create_user(
    bridge_ip: str = env.HUE_BRIDGE_IP_ADDRESS,
    wait_for_link_press: bool = True,
) -> str:
    URL = f"http://{bridge_ip}/api"

    payload = {
        "devicetype": "my_hue_app#python_script",
    }

    if wait_for_link_press:
        input("Press the link button on the Hue Bridge ...")

    try:
        response = requests.post(URL, json=payload)

        if response.status_code == 200:
            result = response.json()
            if "success" in result[0]:
                username = result[0]["success"]["username"]
                logger.info(f"created {username}")
                return username

            logger.error(result)
            return ""

        logger.error(response)
        return ""
    except Exception as e:
        logger.error(e)
        return ""


# hue-2025-03-14-4r9mgh
def list_lights(
    bridge_ip: str = env.HUE_BRIDGE_IP_ADDRESS,
    username: str = env.HUE_BRIDGE_USERNAME,
    verbose: bool = False,
) -> Tuple[bool, Dict]:
    URL = f"http://{bridge_ip}/api/{username}"

    response = requests.get(URL)

    if response.status_code != 200:
        logger.error(response)
        return False, {}

    list_of_lights = response.json()["lights"]
    logger.info(
        "found {} light(s): {}".format(
            len(list_of_lights),
            ", ".join(list_of_lights.keys()),
        )
    )
    if verbose:
        for light_id, light_info in list_of_lights.items():
            logger.info(f"#{light_id}: {light_info}")

    return True, list_of_lights


# hue-2025-03-13-1xjr1z
def set_light_color(
    light_id: str,
    hue: int,  # 0 to 65535
    saturation: int,  # 0 to 254
    bridge_ip: str = env.HUE_BRIDGE_IP_ADDRESS,
    username: str = env.HUE_BRIDGE_USERNAME,
    verbose: bool = False,
) -> bool:
    logger.info(
        "{}.set_light_color({}@{}:{}) -> hue=0x{:x}, saturation=0x{:x}".format(
            NAME,
            username,
            bridge_ip,
            light_id,
            hue,
            saturation,
        )
    )

    # Construct the API endpoint URL
    url = f"http://{bridge_ip}/api/{username}/lights/{light_id}/state"

    # Prepare the payload with the desired hue and saturation
    payload = {
        "hue": hue,
        "sat": saturation,
    }

    response = requests.put(url, json=payload)

    # https://chat.openai.com/c/6deb94d0-826a-48de-b5ef-f7d8da416c82
    # response.raise_for_status()
    if response.status_code // 100 != 2:
        logger.error(response)
        return False

    if verbose:
        logger.info(response.json())

    return True


def test(
    colormap: int = cv2.COLORMAP_HOT,
    light_id: str = "all",
    interval: float = env.HUE_TEST_DEFAULT_INTERVAL,
    bridge_ip: str = env.HUE_BRIDGE_IP_ADDRESS,
    username: str = env.HUE_BRIDGE_USERNAME,
    verbose: bool = False,
) -> bool:
    if not light_id:
        light_id = "all"

    logger.info(
        "{}.test({}@{}:{}) @ colormap #{}, interval={} s".format(
            NAME,
            username,
            bridge_ip,
            light_id,
            colormap,
            interval,
        )
    )

    list_of_lights: List[str]
    if light_id == "all":
        success, dict_of_lights = list_lights(
            bridge_ip=bridge_ip,
            username=username,
            verbose=verbose,
        )
        if not success:
            return success
        list_of_lights = list(dict_of_lights.keys())
    else:
        list_of_lights = [light_id]

    list_of_hue_values = get_hue_values(
        colormap=colormap,
        length=len(list_of_lights),
    )
    list_of_hue_values = list_of_hue_values + list_of_hue_values

    hue_offset: int = 0
    try:
        while True:
            logger.info(f"hue_offset={hue_offset}")

            for light_index in tqdm(range(len(list_of_lights))):
                set_light_color(
                    light_id=list_of_lights[light_index],
                    hue=list_of_hue_values[hue_offset + light_index],
                    saturation=random.randint(1, env.HUE_MAX_SATURATION),
                    bridge_ip=bridge_ip,
                    username=username,
                    verbose=verbose,
                )

                sleep(interval)

            hue_offset = (hue_offset + 1) % len(list_of_lights)

    except KeyboardInterrupt:
        logger.info("Ctrl+C detected.")

    return True



from blueness import module

from blue_options.logger import log_list

from blue_assistant import NAME
from blue_assistant.web.crawl import crawl_list_of_urls
from blue_assistant.script.repository.base.classes import BaseScript
from blue_assistant.logger import logger


NAME = module.name(__file__, NAME)


def web_crawl(
    script: BaseScript,
    node_name: str,
    use_cache: bool,
) -> bool:
    logger.info(f"{NAME}: {script} @ {node_name} ...")

    seed_url_var_name = script.nodes[node_name].get("seed_urls", "")
    if not isinstance(seed_url_var_name, str):
        logger.error(f"{node_name}: seed_urls must be a string.")
        return False
    if not seed_url_var_name:
        logger.error(f"{node_name}: seed_urls not found.")
        return False

    # to allow both :::<var-name> and <var-name> - for convenience :)
    if seed_url_var_name.startswith(":::"):
        seed_url_var_name = seed_url_var_name[3:].strip()

    if seed_url_var_name not in script.vars:
        logger.error(f"{node_name}: {seed_url_var_name}: seed_urls not found in vars.")
        return False
    seed_urls = script.vars[seed_url_var_name]
    log_list(logger, "using", seed_urls, "seed url(s)")

    visited_urls = crawl_list_of_urls(
        seed_urls=seed_urls,
        object_name=script.object_name,
        max_iterations=script.nodes[node_name]["max_iterations"],
        use_cache=use_cache,
    )

    script.nodes[node_name]["visited_urls"] = visited_urls
    script.nodes[node_name]["output"] = "TBA"

    return True



from typing import Dict, Any, List
import requests
from bs4 import BeautifulSoup, XMLParsedAsHTMLWarning
from urllib.parse import urljoin
import re
import warnings

from blueness import module
from blue_options.logger import log_long_text, log_list

from blue_assistant import NAME
from blue_assistant.logger import logger

warnings.filterwarnings("ignore", category=XMLParsedAsHTMLWarning)

NAME = module.name(__file__, NAME)


def fetch_links_and_text(
    url: str,
    verbose: bool = False,
) -> Dict[str, Any]:
    try:
        response = requests.get(url, timeout=5)
    except Exception as e:
        logger.warning(e)
        return {}

    if response.status_code != 200:
        logger.error(response)
        return {}

    content_type = response.headers.get("Content-Type", "")
    logger.info(f"content-type: {content_type}")

    list_of_urls: List[str] = []
    list_of_ignored_urls: List[str] = []
    text = ""

    if not any(
        thing in content_type
        for thing in [
            "pdf",
            "xml",
        ]
    ):
        soup = BeautifulSoup(response.text, "html.parser")

        for a_tag in soup.find_all("a", href=True):
            a_url = urljoin(url, a_tag["href"])

            if a_url.startswith(url):
                if url not in list_of_urls:
                    logger.info(f"+= {a_url}")
                    list_of_urls += [a_url]
                continue

            if a_url not in list_of_ignored_urls:
                list_of_ignored_urls += [a_url]
                if verbose:
                    logger.info(f"ignored: {a_url}")

        text = soup.get_text(separator=" ", strip=True)

    # remove non-ASCII characters
    text = re.sub(r"[^\x20-\x7E]+", "", text)
    for thing in ["\r", "\n", "\t"]:
        text = text.replace(thing, " ")
    text = re.sub(r"\s+", " ", text).strip()

    if verbose:
        log_list(logger, "fetched", list_of_urls, "url(s)")
        log_list(logger, "ignored", list_of_ignored_urls, "url(s)")
        log_long_text(logger, text)
    else:
        logger.info(
            "{} url(s) collected, {} url(s) ignored, text: {:,} char(s).".format(
                len(list_of_urls),
                len(list_of_ignored_urls),
                len(text),
            )
        )

    return {
        "url": url,
        "content_type": content_type,
        "list_of_ignored_urls": list_of_ignored_urls,
        "list_of_urls": list_of_urls,
        "text": text,
    }



from blue_objects import file, path

from blue_assistant.script.repository.generic.classes import GenericScript


class OrbitalDataExplorerScript(GenericScript):
    name = path.name(file.path(__file__))



import argparse
import cv2

from blueness import module
from blueness.argparse.generic import sys_exit

from blue_assistant import NAME
from blue_assistant import env
from blue_assistant.script.repository.hue.api import (
    create_user,
    list_lights,
    set_light_color,
    test,
)
from blue_assistant.logger import logger

NAME = module.name(__file__, NAME)

parser = argparse.ArgumentParser(NAME)
parser.add_argument(
    "task",
    type=str,
    help="create_user | list | set | test",
)
parser.add_argument(
    "--bridge_ip",
    type=str,
    default=env.HUE_BRIDGE_IP_ADDRESS,
)
parser.add_argument(
    "--username",
    type=str,
    default=env.HUE_BRIDGE_USERNAME,
    help="aka API key",
)
parser.add_argument(
    "--light_id",
    type=str,
    default="",
    help="all | <light_id>",
)
parser.add_argument(
    "--hue",
    type=int,
    default=65535,
    help="0 to 65535",
)
parser.add_argument(
    "--saturation",
    type=int,
    default=env.HUE_MAX_SATURATION,
    help=f"0 to {env.HUE_MAX_SATURATION}",
)
parser.add_argument(
    "--verbose",
    type=int,
    default=0,
    help="0 | 1",
)
parser.add_argument(
    "--interval",
    type=float,
    default=env.HUE_TEST_DEFAULT_INTERVAL,
    help="in seconds",
)
parser.add_argument(
    "--colormap",
    type=int,
    default=cv2.COLORMAP_HOT,
    help="//docs.opencv.org/4.x/d3/d50/group__imgproc__colormap.html",
)
args = parser.parse_args()

success = False
if args.task == "create_user":
    username = create_user(
        bridge_ip=args.bridge_ip,
    )
    if username:
        success = True
elif args.task == "list":
    success, _ = list_lights(
        bridge_ip=args.bridge_ip,
        username=args.username,
        verbose=args.verbose == 1,
    )
elif args.task == "set":
    success = set_light_color(
        bridge_ip=args.bridge_ip,
        username=args.username,
        light_id=args.light_id,
        hue=args.hue,
        saturation=args.saturation,
        verbose=args.verbose == 1,
    )
elif args.task == "test":
    success = test(
        bridge_ip=args.bridge_ip,
        username=args.username,
        light_id=args.light_id,
        interval=args.interval,
        colormap=args.colormap,
        verbose=args.verbose == 1,
    )
else:
    success = None

sys_exit(logger, NAME, args.task, success)



from typing import List
import cv2

from blue_options.terminal import show_usage, xtra

from blue_assistant import env


def help_create_user(
    tokens: List[str],
    mono: bool,
) -> str:
    options = xtra("dryrun", mono=mono)

    args = [
        f"[--bridge_ip <{env.HUE_BRIDGE_IP_ADDRESS}>]",
    ]

    return show_usage(
        [
            "@hue",
            "create_user",
            f"[{options}]",
        ]
        + args,
        "create a hue user.",
        mono=mono,
    )


def help_list(
    tokens: List[str],
    mono: bool,
) -> str:
    options = xtra("dryrun", mono=mono)

    args = [
        f"[--bridge_ip <{env.HUE_BRIDGE_IP_ADDRESS}>]",
        f"[--username <{env.HUE_BRIDGE_USERNAME}>]",
        "[--verbose 1]",
    ]

    return show_usage(
        [
            "@hue",
            "list",
            f"[{options}]",
        ]
        + args,
        "list hue lights.",
        mono=mono,
    )


def help_set(
    tokens: List[str],
    mono: bool,
) -> str:
    options = xtra("dryrun", mono=mono)

    args = [
        f"[--bridge_ip <{env.HUE_BRIDGE_IP_ADDRESS}>]",
        f"[--username <{env.HUE_BRIDGE_USERNAME}>]",
        "[--light_id <light_id>]",
        "[--hue <65535>]",
        f"[--saturation <{env.HUE_MAX_SATURATION}>]",
        "[--verbose 1]",
    ]

    return show_usage(
        [
            "@hue",
            "set",
            f"[{options}]",
        ]
        + args,
        "set hue lights.",
        mono=mono,
    )


def help_test(
    tokens: List[str],
    mono: bool,
) -> str:
    options = xtra("dryrun", mono=mono)

    args = [
        f"[--bridge_ip <{env.HUE_BRIDGE_IP_ADDRESS}>]",
        f"[--username <{env.HUE_BRIDGE_USERNAME}>]",
        "[--light_id all | <light_id>]",
        f"[--interval <{env.HUE_TEST_DEFAULT_INTERVAL:.2f}>]",
        f"[--colormap <{cv2.COLORMAP_HOT}>]",
        "[--verbose 1]",
    ]

    return show_usage(
        [
            "@hue",
            "test",
            f"[{options}]",
        ]
        + args,
        "test hue.",
        {
            "colormap: https://docs.opencv.org/4.x/d3/d50/group__imgproc__colormap.html": [],
        },
        mono=mono,
    )


help_functions = {
    "create_user": help_create_user,
    "list": help_list,
    "set": help_set,
    "test": help_test,
}



import os

from blue_options.help.functions import get_help
from blue_objects import file, README

from blue_assistant import NAME, VERSION, ICON, REPO_NAME
from blue_assistant.help.functions import help_functions


items = README.Items(
    [
        {
            "name": "orbital-data-explorer",
            "url": "./blue_assistant/script/repository/orbital_data_explorer",
            "marquee": "https://github.com/kamangir/assets/blob/main/blue-assistant/orbital-data-explorer.png?raw=true",
            "description": "Access to the [Orbital Data Explorer](https://ode.rsl.wustl.edu/). 🔥",
        },
        {
            "name": "🌀 blue script",
            "marquee": "https://github.com/kamangir/assets/raw/main/blue-plugin/marquee.png?raw=true",
            "description": "A minimal AI DAG interface.",
            "url": "./blue_assistant/script/",
        },
        {
            "name": "@hue",
            "url": "./blue_assistant/script/repository/hue",
            "marquee": "https://github.com/kamangir/assets/raw/main/blue-assistant/20250314_143702.jpg?raw=true",
            "description": '"send a color command to the Hue LED lights in my apartment."',
        },
        {
            "name": "blue-amo",
            "url": "./blue_assistant/script/repository/blue_amo/README.md",
            "marquee": "https://github.com/kamangir/assets/raw/main/blue-amo-2025-02-03-nswnx6/stitching_the_frames-2.png?raw=true",
            "description": "Story development and visualization.",
        },
    ]
)


def build():
    return all(
        README.build(
            items=readme.get("items", []),
            cols=readme.get("cols", 3),
            path=os.path.join(file.path(__file__), readme["path"]),
            ICON=ICON,
            NAME=NAME,
            VERSION=VERSION,
            REPO_NAME=REPO_NAME,
            help_function=lambda tokens: get_help(
                tokens,
                help_functions,
                mono=True,
            ),
        )
        for readme in [
            {
                "items": items,
                "cols": 2,
                "path": "..",
            },
            {"path": "script/repository/blue_amo"},
            #
            {"path": "script/repository/orbital_data_explorer/docs/round-1.md"},
            {"path": "script/repository/orbital_data_explorer/docs"},
            #
            {"path": "script/repository/hue/docs/round-1.md"},
            {"path": "script/repository/hue/docs"},
            #
            {"path": "script/"},
            {"path": "web/"},
        ]
    )



from typing import List

from blue_options.terminal import show_usage, xtra


def help_crawl(
    tokens: List[str],
    mono: bool,
) -> str:
    options = xtra("cache,~download,dryrun,~upload", mono=mono)

    args = [
        "[--max_iterations <100000>]",
    ]

    return show_usage(
        [
            "@web",
            "crawl",
            f"[{options}]",
            "<url-1>+<url-2>+<url-3>",
            "[-|<object-name>]",
        ]
        + args,
        "crawl the urls.",
        mono=mono,
    )


def help_fetch(
    tokens: List[str],
    mono: bool,
) -> str:
    options = xtra("dryrun,~upload", mono=mono)

    return show_usage(
        [
            "@web",
            "fetch",
            f"[{options}]",
            "<url>",
            "[-|<object-name>]",
        ],
        "fetch <url>.",
        mono=mono,
    )


help_functions = {
    "crawl": help_crawl,
    "fetch": help_fetch,
}



import argparse

from blueness import module
from blueness.argparse.generic import sys_exit
from blue_options.logger import log_dict
from blue_objects.metadata import post_to_object

from blue_assistant import NAME
from blue_assistant.web import crawl_list_of_urls, fetch_links_and_text
from blue_assistant.logger import logger

NAME = module.name(__file__, NAME)

parser = argparse.ArgumentParser(NAME)
parser.add_argument(
    "task",
    type=str,
    help="crawl | fetch",
)
parser.add_argument(
    "--max_iterations",
    type=int,
    default=10,
)
parser.add_argument(
    "--verbose",
    type=int,
    default=0,
    help="0 | 1",
)
parser.add_argument(
    "--url",
    type=str,
)
parser.add_argument(
    "--seed_urls",
    type=str,
)
parser.add_argument(
    "--object_name",
    type=str,
)
parser.add_argument(
    "--use_cache",
    type=int,
    default=0,
    help="0 | 1",
)
args = parser.parse_args()

success = False
if args.task == "crawl":
    success = True
    dict_of_urls = crawl_list_of_urls(
        seed_urls=args.seed_urls.split("+"),
        object_name=args.object_name,
        max_iterations=args.max_iterations,
        use_cache=args.use_cache == 1,
    )

    if args.verbose == 1:
        log_dict(logger, "crawled", dict_of_urls, "url(s)")
elif args.task == "fetch":
    summary = fetch_links_and_text(
        url=args.url,
        verbose=True,
    )

    success = post_to_object(
        args.object_name,
        NAME.replace(".", "-"),
        summary,
    )
else:
    success = None

sys_exit(logger, NAME, args.task, success)



from typing import List
import numpy as np
import cv2


# https://chatgpt.com/c/67d4c06c-2168-8005-9d8b-4b6c9848957e
def get_hue_values(
    colormap: int = cv2.COLORMAP_HOT,
    length: int = 255,
) -> List[int]:
    # Create a gradient from 0 to 255
    gradient = np.linspace(0, 255, length).astype("uint8")
    gradient = np.repeat(gradient[np.newaxis, :], 1, axis=0)

    # Apply the colormap
    color_mapped_image = cv2.applyColorMap(gradient, colormap)

    # Convert BGR to HSV
    hsv_image = cv2.cvtColor(color_mapped_image, cv2.COLOR_BGR2HSV)

    return (hsv_image[0, :, 0] * (65535 / 179)).astype(int).tolist()



import re


def url_to_filename(
    url: str,
    max_length: int = 255,
) -> str:
    # Remove the URL scheme (http://, https://)
    filename = re.sub(r"^https?://", "", url)

    # Replace unwanted characters with an underscore
    filename = re.sub(r"[^\w\s-]", "_", filename)

    # Replace slashes with a hyphen to preserve some structure
    filename = re.sub(r"\/", "-", filename)

    # Replace spaces with underscores
    filename = filename.replace(" ", "_")

    # Ensure the filename length is not too long
    if len(filename) > max_length:
        filename = filename[:max_length]

    return filename



from blue_options.env import load_config, load_env, get_env

load_env(__name__)
load_config(__name__)


BLUE_ASSISTANT_TEXT_DEFAULT_MODEL = get_env("BLUE_ASSISTANT_TEXT_DEFAULT_MODEL")

BLUE_ASSISTANT_TEXT_MAX_TOKENS = get_env("BLUE_ASSISTANT_TEXT_MAX_TOKENS", 2000)

BLUE_ASSISTANT_IMAGE_DEFAULT_MODEL = get_env("BLUE_ASSISTANT_IMAGE_DEFAULT_MODEL")

BLUE_ASSISTANT_IMAGE_DEFAULT_QUALITY = get_env("BLUE_ASSISTANT_IMAGE_DEFAULT_QUALITY")

BLUE_ASSISTANT_IMAGE_DEFAULT_SIZE = get_env("BLUE_ASSISTANT_IMAGE_DEFAULT_SIZE")

HUE_BRIDGE_IP_ADDRESS = get_env("HUE_BRIDGE_IP_ADDRESS")

HUE_BRIDGE_USERNAME = get_env("HUE_BRIDGE_USERNAME")
HUE_TEST_DEFAULT_INTERVAL = get_env("HUE_TEST_DEFAULT_INTERVAL", 0.01)
HUE_MAX_SATURATION = get_env("HUE_MAX_SATURATION", 254)



from typing import List, Dict

from blueness import module
from blue_options.logger import log_dict, log_list
from blue_objects import file
from blue_objects import objects
from blue_objects.metadata import get_from_object, post_to_object

from blue_assistant import NAME
from blue_assistant.web.fetch import fetch_links_and_text
from blue_assistant.web.functions import url_to_filename
from blue_assistant.logger import logger

NAME = module.name(__file__, NAME)


def crawl_list_of_urls(
    seed_urls: List[str],
    object_name: str,
    max_iterations: int = 10,
    use_cache: bool = False,
    verbose: bool = False,
) -> Dict[str, str]:
    logger.info(
        "{}.crawl_list_of_urls({}): {} -{}> {}".format(
            NAME,
            len(seed_urls),
            ", ".join(seed_urls),
            "use-cache-" if use_cache else "",
            object_name,
        )
    )

    crawl_cache: Dict[str, str] = {}
    queue: List[str] = [url for url in seed_urls]

    if use_cache:
        crawl_cache = get_from_object(
            object_name,
            "crawl_cache",
            {},
        )
        log_dict(logger, "loaded cache:", crawl_cache, "url(s)")

        queue += get_from_object(
            object_name,
            "crawl_queue",
            [],
        )

    log_list(logger, "queue:", queue, "url(s)")

    iteration: int = 0
    while queue:
        url = queue[0]
        queue = queue[1:]

        logger.info(
            "{} {} ...".format(
                "✅ " if url in crawl_cache else "🔗 ",
                url,
            )
        )
        if url in crawl_cache:
            continue

        url_summary = fetch_links_and_text(
            url=url,
            verbose=verbose,
        )
        content_type = url_summary.get("content_type", "unknown")

        if use_cache and "html" in content_type:
            file.save_yaml(
                filename=objects.path_of(
                    object_name=object_name,
                    filename="crawl_summary_cache/{}.yaml".format(url_to_filename(url)),
                ),
                data=url_summary,
            )

        crawl_cache[url] = content_type
        if "list_of_urls" in url_summary:
            queue = list(
                set(
                    queue
                    + [
                        url
                        for url in url_summary["list_of_urls"]
                        if url not in crawl_cache.keys()
                    ]
                )
            )

        iteration += 1
        if max_iterations != -1 and iteration >= max_iterations:
            logger.warning(f"max iteration of {max_iterations} reached.")
            break

    if queue:
        logger.warning(f"queue: {len(queue)}")

    if use_cache:
        post_to_object(
            object_name,
            "crawl_cache",
            crawl_cache,
        )

        post_to_object(
            object_name,
            "crawl_queue",
            queue,
        )

    log_dict(logger, "crawled", crawl_cache, "url(s)")
    log_list(logger, "queue:", queue, "url(s)")

    return crawl_cache
